/************************************************************************/
/* xad.c								*/
/*	This file contains the xad routines. These routines allow an	*/
/* application to easily set up an X Window and display the contents	*/
/* of an array.								*/
/*	This was written in a hurry...please report bugs, etc. (Thanx!)	*/
/*									*/
/* Chris Faigle								*/
/* Northeast Parallel Architectures Center				*/
/* cfaigle@npac.syr.edu							*/
/*									*/
/* Thanx: Dave Koester							*/
/*									*/
/* History:								*/
/*	June 5, 1993	CTF	Started.				*/
/*									*
/************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "xad.h"

#ifdef	NO_ANSI_PROTOTYPES
int	xad_create_colormap();
int	xad_create_eight_bit_colormap();
void	xad_create_hsv_colormap();
void	xad_create_rgb_colormap();
void	xad_hsv_to_rgb();
#else
int	xad_create_colormap(int);
int	xad_create_eight_bit_colormap(int);
void	xad_create_hsv_colormap(XColor *);
void	xad_create_rgb_colormap(XColor *);
void	xad_hsv_to_rgb(float *, float *, float *, float, float, float);
#endif


/* GLOBALS FOLLOW:			*/
	Display		*theDisplay=(Display *)NULL;
    	Window		theWindow=(Window)NULL;
	GC		theGC;
	int		theScreen;
	int		theDepth;
static int		arrayDepth;
static Colormap		theColormap=(Colormap)NULL;
static long		theForegroundPixel;
static long		theBackgroundPixel;
static int		height=0;
static int		width=0;

#define	BORDER_WIDTH	2


/************************************************************************/
/* Function:	xad_open_window						*/
/* Parameters:	int	x	X Position of window			*/
/*		int	y	Y Position of window			*/
/*		int	h	Height of window			*/
/*		int	w	Width of window				*/
/*		int	black	TRUE if Black Background		*/
/*		int	type	see xad.h for valid types		*/
/*		char	*name	Name of the window			*/
/************************************************************************/

#ifdef	NO_ANSI_PROTOTYPES
int xad_open_window(x,y,h,w,type,name)
int	x;
int	y;
int	h;
int	w;
int	type;
char	*name;
#else	NO_ANSI_PROTOTYPES
int xad_open_window(int x, int y, int h, int w, int type, char *name)
#endif	NO_ANSI_PROTOTYPES
{
	XSetWindowAttributes	theWindowAttributes;
	XSizeHints		theSizeHints;
	unsigned long		theWindowMask;
	XGCValues		theGCValues;
        XEvent                  the_event;
	int			min_depth;
	int			flag;
	static int		first_time=TRUE;

	switch(type)
	{
		case	XAD_WT_8_HSV:
		case	XAD_WT_8_RGB:
		case	XAD_WT_8_GREY:
		case	XAD_WT_8_RED:
		case	XAD_WT_8_GREEN:
		case	XAD_WT_8_BLUE:
			min_depth=8;
			break;
		case	XAD_WT_1_BW:
		case	XAD_WT_1_WB:
			min_depth=1;
			break;
		default:
			return(FALSE);
	}
	if(theWindow!=(Window)NULL)
	{
		return(FALSE);
	}
	if(first_time)
	{
		first_time=FALSE;
		theDisplay=XOpenDisplay(NULL);
		if(theDisplay!=NULL)
		{
			theScreen=DefaultScreen(theDisplay);
			theDepth=DefaultDepth(theDisplay,theScreen);
		}
		else
		{
			fprintf(stderr,"Could not open display!\n");
			return(FALSE);
		}
	}

	if(theDepth<min_depth)
	{
		return(FALSE);
	}
	arrayDepth=min_depth;

	switch(type)
	{
		case	XAD_WT_1_BW:
			theForegroundPixel=BlackPixel(theDisplay,theScreen);
			theBackgroundPixel=WhitePixel(theDisplay,theScreen);
			break;
		default:
			theForegroundPixel=WhitePixel(theDisplay,theScreen);
			theBackgroundPixel=BlackPixel(theDisplay,theScreen);
			break;
	}

	width=w;
	height=h;

	theWindowAttributes.border_pixel=BlackPixel(theDisplay,theScreen);
	theWindowAttributes.background_pixel=theBackgroundPixel;
	theWindowAttributes.backing_store=Always;
	theWindowMask=CWBackPixel | CWBorderPixel | CWBackingStore;
	if(xad_create_colormap(type)!=TRUE)
	{
		fprintf(stderr,"Error setting colormap\n");
		return(FALSE);
	}
#ifdef	WAHOO
	if(theColormap!=(Colormap) NULL)
	{
		printf("Setting Colormap\n");
		theWindowAttributes.colormap=theColormap;
		theWindowMask|=CWColormap;
	}
#endif	WAHOO
	theWindow=XCreateWindow(theDisplay,RootWindow(theDisplay,theScreen),
					x,y,width,height,BORDER_WIDTH,
					theDepth,InputOutput,CopyFromParent,
					theWindowMask,&theWindowAttributes);


	XStoreName(theDisplay,theWindow,name);

	theSizeHints.flags=USPosition | USSize | PMinSize | PMaxSize;
	theSizeHints.x=x;
	theSizeHints.y=y;
	theSizeHints.width=width;
	theSizeHints.height=height;
	theSizeHints.min_width=width;
	theSizeHints.max_width=width;
	theSizeHints.min_height=height;
	theSizeHints.max_height=height;
	XSetNormalHints(theDisplay,theWindow,&theSizeHints);



	/* Now creat the Graphics Context	*/
	theGC=XCreateGC(theDisplay,theWindow,(unsigned long)0,&theGCValues);

	if(theGC==0)
	{
		fprintf(stderr,"COULD NOT CREATE GC!!\n");
		return(FALSE);
	}
	else
	{
		XSetForeground(theDisplay,theGC,theForegroundPixel);
		XSetBackground(theDisplay,theGC,theBackgroundPixel);
	}
	XMapWindow(theDisplay,theWindow);
	/* Wait for event:	*/
        XSelectInput(theDisplay,theWindow,(long)0xFFFFFF);
        XMapWindow(theDisplay,theWindow);
        XFlush(theDisplay);
        flag=FALSE;
        while(!flag)
        {
                XNextEvent(theDisplay,&the_event);
                switch(the_event.type)
                {
                        case    Expose:
                                flag=TRUE;
                                break;
                        case    ClientMessage:
                                flag=TRUE;
                                break;
                        case    MappingNotify:
                                flag=TRUE;
                                break;
                        default:
                                flag=TRUE;
                                break;
                }
        }
/*      printf("Done with opening the window...\n");    */
        XSelectInput(theDisplay,theWindow,NoEventMask);

	if(theColormap!=(Colormap)NULL)
	{
		XSetWindowColormap(theDisplay,theWindow,theColormap);
	}
	XFlush(theDisplay);
	XSync(theDisplay,False);
	return(TRUE);
}


/************************************************************************/
/* Function:	xad_open_window						*/
/* Parameters:	NONE							*/
/************************************************************************/
int xad_close_window()
{
	if(theWindow!=(Window)NULL)
	{
		XFreeGC(theDisplay,theGC);
		XDestroyWindow(theDisplay,theWindow);
		theWindow=(Window)NULL;
	}
	if(theColormap!=(Colormap)NULL)
	{
		XFreeColormap(theDisplay,theColormap);
		theColormap=(Colormap)NULL;
	}
	return(TRUE);
}


/************************************************************************/
/* Function:	xad_dump_array						*/
/* Parameters:	void	*array	Data to be dumped			*/
/*		int	depth	#of bits per element			*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
int xad_dump_array(array)
char	*array;
#else	NO_ANSI_PROTOTYPES
int xad_dump_array(char *array)
#endif	NO_ANSI_PROTOTYPES
{
	XImage	*theImage;
	Pixmap	thePixmap;

	if(theWindow==(Window)NULL)
	{
		return(FALSE);
	}
	if(arrayDepth==8)
	{
		/* Now dump the array into the window	*/
		theImage=XCreateImage(theDisplay,
				DefaultVisual(theDisplay,theScreen),
				theDepth,ZPixmap,0,array,width,height,8,width);

		if(theImage==(XImage *)NULL)
		{
			fprintf(stderr,"Could not allocate image\n");
			return(FALSE);
		}
		XPutImage(theDisplay,theWindow,theGC,theImage,0,0,0,0,
				width,height);
		XDestroyImage(theImage);
	}
	else
	{
		thePixmap=XCreatePixmapFromBitmapData(theDisplay,theWindow,
				array,width,height,theForegroundPixel,
				theBackgroundPixel,theDepth);

		if(thePixmap==(Pixmap)NULL)
		{
			fprintf(stderr,"Could not allocate pixmap\n");
			return(FALSE);
		}
		XCopyArea(theDisplay,thePixmap,theWindow,theGC,0,0,
				width,height,0,0);

		XFreePixmap(theDisplay,thePixmap);
	}
	XFlush(theDisplay);
	XSync(theDisplay,False);
	return(TRUE);	

}


/************************************************************************/
/* Function:	xad_make_colormap					*/
/* Parameters:	int	type	Type of array/window			*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
int xad_create_colormap(type)
int	type;
#else	NO_ANSI_PROTOTYPES
int xad_create_colormap(int type)
#endif	NO_ANSI_PROTOTYPES
{
	switch(type)
	{
		case	XAD_WT_8_HSV:
		case	XAD_WT_8_RGB:
		case	XAD_WT_8_GREY:
		case	XAD_WT_8_RED:
		case	XAD_WT_8_GREEN:
		case	XAD_WT_8_BLUE:
			return(xad_create_eight_bit_colormap(type));
			break;
		case	XAD_WT_1_BW:
		case	XAD_WT_1_WB:
			return(TRUE);
			break;
		default:
			return(FALSE);
	}
}


/************************************************************************/
/* Function:	xad_create_eight_bit_colormap				*/
/* Parameters:	int	type	Type of array/window			*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
int xad_create_eight_bit_colormap(type)
int	type;
#else	NO_ANSI_PROTOTYPES
int xad_create_eight_bit_colormap(int type)
#endif	NO_ANSI_PROTOTYPES
{
	int		loop1;
	XColor		color_cells[256];

	theColormap=XCreateColormap(theDisplay,RootWindow(theDisplay,
			theScreen),DefaultVisual(theDisplay,theScreen),
			AllocAll);

	for(loop1=0;loop1<256;++loop1)
	{
		color_cells[loop1].flags=DoRed | DoBlue | DoGreen;
		color_cells[loop1].pad=0;
		color_cells[loop1].pixel=(long)loop1;
	}
	switch(type)
	{
		case	XAD_WT_8_HSV:
			xad_create_hsv_colormap(color_cells);
			break;
		case	XAD_WT_8_RGB:
			xad_create_rgb_colormap(color_cells);
			break;
		case	XAD_WT_8_GREY:
			for(loop1=0;loop1<256;++loop1)
			{
				color_cells[loop1].red=loop1*256;
				color_cells[loop1].green=loop1*256;
				color_cells[loop1].blue=loop1*256;
			}			
			break;
		case	XAD_WT_8_RED:
			for(loop1=0;loop1<256;++loop1)
			{
				color_cells[loop1].red=loop1*256;
				color_cells[loop1].green=0;
				color_cells[loop1].blue=0;
			}			
			break;
		case	XAD_WT_8_GREEN:
			for(loop1=0;loop1<256;++loop1)
			{
				color_cells[loop1].red=0;
				color_cells[loop1].green=loop1*256;
				color_cells[loop1].blue=0;
			}			
			break;
		case	XAD_WT_8_BLUE:
			for(loop1=0;loop1<256;++loop1)
			{
				color_cells[loop1].red=0;
				color_cells[loop1].green=0;
				color_cells[loop1].blue=loop1*256;
			}			
			break;
	}
	XStoreColors(theDisplay,theColormap,color_cells,256);
	return(TRUE);
}


/************************************************************************/
/* Function:	xad_create_hsv_colormap					*/
/* Parameters:	XColor	*color_cells	Array of colors to be set	*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
void xad_create_hsv_colormap(color_cells)
XColor	*color_cells;
#else	NO_ANSI_PROTOTYPES
void xad_create_hsv_colormap(XColor *color_cells)
#endif	NO_ANSI_PROTOTYPES
{
	int	loop1;
	float	red,green,blue;
	float	k=360/255;

	for(loop1=0;loop1<256;++loop1)
	{
						/* s=1.0, v=1.0	*/
		xad_hsv_to_rgb(&red,&green,&blue,k*loop1,1.0,1.0);
		color_cells[loop1].red=(short)red*65535;
		color_cells[loop1].green=(short)green*65535;
		color_cells[loop1].blue=(short)blue*65535;
	}			
}


/************************************************************************/
/* Function:	xad_create_rgb_colormap					*/
/* Parameters:	XColor	*color_cells	Array of colors to be set	*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
void xad_create_rgb_colormap(color_cells)
XColor	*color_cells;
#else	NO_ANSI_PROTOTYPES
void xad_create_rgb_colormap(XColor *color_cells)
#endif	NO_ANSI_PROTOTYPES
{
	int	loop1;
	short	color[3];

	for(loop1=0;loop1<256;++loop1)
	{
		color_cells[loop1].red=((loop1&0x00C0)>>6)*64*240+4096;
		color_cells[loop1].green=((loop1&0x0038)>>3)*32*240+4096;
		color_cells[loop1].blue=(loop1&0x0007)*32*240+4096;
	}			
}


/************************************************************************/
/* Function:	xad_hsv_to_rgb						*/
/* Parameters:	float	*red	Red return value			*/
/*		float	*green	Green return value			*/
/*		float	*blue	Blue return value			*/
/*		float	h	Hue					*/
/*		float 	s	Saturation				*/
/*		float	v	Value					*/
/* This routine is from Foley, VanDam, Feiner & Hughes: Computer	*/
/*	 Graphics: Principles & Practice				*/
/************************************************************************/
#ifdef	NO_ANSI_PROTOTYPES
void	xad_hsv_to_rgb(red, green, blue, h, s, v)
float	*red;
float	*green;
float	*blue;
float	h;
float	s;
float	v;
#else	NO_ANSI_PROTOTYPES
void	xad_hsv_to_rgb(float *red, float *green, float *blue,
				   float h, float s, float v)
#endif	NO_ANSI_PROTOTYPES

{
	int	i;
	float	f;
	float	p;
	float	q;
	float	t;

	h=h/60;
	i=(int)h;
	f=h-i;
	p=v*(1-s);
	q=v*(1-s*f);
	t=v*(1-s*(1-f));
	switch(i)
	{
		case	0:
			*red=v;
			*blue=t;
			*green=p;
			break;
		case	1:
			*red=q;
			*blue=v;
			*green=p;
			break;
		case	2:
			*red=p;
			*blue=v;
			*green=t;
			break;
		case	3:
			*red=p;
			*blue=q;
			*green=v;
			break;
		case	4:
			*red=t;
			*blue=p;
			*green=v;
			break;
		case	5:
			*red=v;
			*blue=p;
			*green=q;
			break;
	}
}
